import math
from decimal import getcontext
from fractions import Fraction

import pytest

from isqx import (
    CELSIUS,
    DAY,
    HOUR,
    M_PERS,
    MIN,
    RAD,
    A,
    BaseUnit,
    CompositionError,
    Dimensionless,
    DimensionMismatchError,
    Exp,
    KindMismatchError,
    LazyProduct,
    M,
    Mul,
    S,
    Scaled,
    Tagged,
    convert,
    dimension,
    simplify,
)
from isqx.aerospace import GEOMETRIC_ALTITUDE, GEOPOTENTIAL_ALTITUDE
from isqx.usc import FT

#
# OpsMixin
#


def test_op_mixin() -> None:
    # pow
    assert M**2 == Exp(M, 2)
    assert M ** Fraction(1, 2) == Exp(M, Fraction(1, 2))

    # mul expr * expr and flattening
    assert M * S == Mul((M, S))
    assert S * A * M == Mul((S, A, M))
    assert S * (A * M) == Mul((S, A, M))
    # rmul factor * expr, mul expr * factor
    assert 2 * M == Scaled(M, 2)
    assert M * 2 == Scaled(M, 2)

    # truediv expr / expr, expr / factor
    assert M / 2 == Scaled(M, LazyProduct(((2, -1),)))
    assert M / S == Mul((M, Exp(S, -1)))

    assert (M * 2) / S == Mul((Scaled(M, 2), Exp(S, -1)))
    assert 3 * (M * 2) == Scaled(Scaled(M, 2), 3)


#
# Exp
#


def test_exp_invalid() -> None:
    with pytest.raises(CompositionError, match="zero"):
        _u1 = Exp(M, 0)


def test_exp_eq() -> None:
    assert M**2 == Exp(M, Fraction(2))


def test_exp_dimension() -> None:
    assert dimension(M**2) == Exp(dimension(M), 2)
    assert dimension((M**2) ** 3) == Exp(Exp(dimension(M), 2), 3)


def test_exp_simplify() -> None:
    # distribute exponent
    expr0s = simplify((M**2) ** Fraction(1, 2))
    assert isinstance(expr0s, BaseUnit)
    assert expr0s == M

    expr0s = simplify((M**2) ** 3)
    assert isinstance(expr0s, Exp)
    assert expr0s == Exp(M, 6)


#
# Mul
#


def test_mul_invalid() -> None:
    from isqx import MixedKindError

    with pytest.raises(CompositionError, match="empty"):
        _u0 = Mul(tuple())
    with pytest.raises(MixedKindError):
        _u1 = Mul((M, dimension(M)))


FT_PERMIN = FT * MIN**-1


def test_mul_dimension() -> None:
    assert dimension(M_PERS) == Mul((dimension(M), Exp(dimension(S), -1)))


def test_mul_simplify_basic() -> None:
    # cancel terms
    expr1s = simplify(M * M**-1)
    assert isinstance(expr1s, Dimensionless)

    # distribute inner
    expr2s = simplify(M_PERS**2)
    assert isinstance(expr2s, Mul)
    assert expr2s.terms == (M**2 * S**-2).terms

    # combine terms with same base
    expr_s = simplify(M * S**-1 * M**2 * S**-2)
    assert isinstance(expr_s, Mul)
    assert expr_s.terms == (M**3 * S**-3).terms

    # return lone term if it is raised to power of one
    expr_s = simplify(Mul((M**1,)))
    assert isinstance(expr_s, BaseUnit)
    assert expr_s == M


def test_mul_simplify_nested() -> None:
    assert isinstance(simplify((M_PERS * M_PERS**-1)), Dimensionless)


def test_mul_simplify_ordering() -> None:
    PERSM = simplify(S**-1 * M)
    assert isinstance(PERSM, Mul)
    assert PERSM.terms == M_PERS.terms


#
# Scaled
#
CTX_DEFAULT = getcontext()


def test_scaled_dimension() -> None:
    assert dimension(FT) == dimension(M)


def test_scaled_simplify_nested() -> None:
    from isqx import YEAR

    expr3s = simplify(YEAR)
    assert isinstance(expr3s, Scaled)
    assert expr3s.reference == S
    assert isinstance(expr3s.factor, LazyProduct)
    assert expr3s.factor.to_exact(ctx=CTX_DEFAULT) == 86400 * 365.25


def test_scaled_simplify_mixed() -> None:
    expr4s = simplify((M * 2) ** 3 * (S * 3) ** 2)
    assert isinstance(expr4s, Scaled)
    assert expr4s.reference == (M**3 * S**2)
    assert isinstance(expr4s.factor, LazyProduct)
    assert expr4s.factor.to_exact(ctx=CTX_DEFAULT) == 2**3 * 3**2

    expr4s = simplify(FT_PERMIN)
    assert isinstance(expr4s, Scaled)
    assert expr4s.reference == M_PERS
    assert isinstance(expr4s.factor, LazyProduct)
    assert expr4s.factor.to_exact(ctx=CTX_DEFAULT) == Fraction(3048, 10000) / 60


def test_scaled_simplify_dimensionless() -> None:
    expr5s = simplify(HOUR * DAY**-1)
    assert isinstance(expr5s, Scaled)
    assert isinstance(expr5s.reference, Dimensionless)
    assert isinstance(expr5s.factor, LazyProduct)
    assert expr5s.factor.to_exact(ctx=CTX_DEFAULT) == Fraction(1, 24)


def test_scaled_simplify_with_lazy_product() -> None:
    from isqx.usc import IN  # -> ft -> m

    result = simplify(Exp(simplify(IN), 2))  # type: ignore

    assert isinstance(result, Scaled)
    assert result.reference == M**2
    assert isinstance(result.factor, LazyProduct)

    assert result.factor.to_approx() == pytest.approx((0.3048 / 12) ** 2)


def test_scaled_simplify_with_lazy_product_multiple_terms() -> None:
    from isqx import PA
    from isqx.usc import PSI

    # defined by lbf -> lbm -> kg and in -> ft -> m
    psi_simplified = simplify(PSI)  # Scaled(simplify(PA), LazyProduct(...))
    assert isinstance(psi_simplified, Scaled)
    assert isinstance(psi_simplified.factor, LazyProduct)
    assert isinstance(psi_simplified.reference, Mul)

    pa_simplified = simplify(PA)
    assert isinstance(pa_simplified, Mul)
    assert set(psi_simplified.reference.terms) == set(pa_simplified.terms)
    PSI_TO_PA_FACTOR = (0.45359237 * 9.80665) / ((0.3048 / 12) ** 2)
    assert psi_simplified.factor.to_approx() == pytest.approx(PSI_TO_PA_FACTOR)

    result = simplify(psi_simplified**2)
    assert isinstance(result, Scaled)
    assert isinstance(result.factor, LazyProduct)
    assert isinstance(result.reference, Mul)

    from isqx import KG

    result_ref_simplified = simplify(result.reference)
    assert isinstance(result_ref_simplified, Mul)
    # P**2 = (F/A)**2
    #      = ((M*L*T**-2) / L**2)**2
    #      = (M * L**-1 * T**-2)**2
    assert set(result_ref_simplified.terms) == set(
        (KG**2 * M**-2 * S**-4).terms
    )
    assert result.factor.to_approx() == pytest.approx(PSI_TO_PA_FACTOR**2)


#
# prefix
#


def test_prefix_invalid() -> None:
    from isqx import KG, KILO, G, W

    KW = KILO * W
    assert isinstance(KW, Scaled)
    assert KW.factor == KILO
    assert KW.reference == W
    with pytest.raises(CompositionError, match="prefixes cannot be applied"):
        _ = KILO * dimension(KG)
    with pytest.raises(CompositionError, match="cannot prefix `kilogram`"):
        _ = KILO * KG
    with pytest.raises(CompositionError, match="apply prefix `kilo` to `gram`"):
        _ = KILO * G
    with pytest.raises(CompositionError, match="prefixes cannot be applied"):
        _ = KILO * M_PERS
    with pytest.raises(
        CompositionError, match="prefixes cannot be applied to `Scaled`"
    ):
        _ = KILO * KW


#
# tagged
#

M_ALT_GEOM: Tagged = GEOMETRIC_ALTITUDE(M)  # type: ignore
M_ALT_GEOP: Tagged = GEOPOTENTIAL_ALTITUDE(M)  # type: ignore


def test_tagged_invalid_construction() -> None:
    with pytest.raises(CompositionError, match="nesting"):
        _ = Tagged(M_ALT_GEOP, ("another_context",))  # type: ignore


def test_tagged_simplify_cancellation() -> None:
    expr_same_ctx = M_ALT_GEOP * M_ALT_GEOP**-1
    simplified_same = simplify(expr_same_ctx)
    assert isinstance(simplified_same, Dimensionless)

    expr_diff_ctx = M_ALT_GEOP * M_ALT_GEOM**-1
    simplified_diff = simplify(expr_diff_ctx)
    assert isinstance(simplified_diff, Mul)  # shouldn't cancel
    assert set(simplified_diff.terms) == set(expr_diff_ctx.terms)


FT_ALT_GEOP: Tagged = GEOPOTENTIAL_ALTITUDE(FT)  # type: ignore


def test_tagged_simplify_propagates_to_reference() -> None:
    simplified = simplify(FT_ALT_GEOP)
    assert isinstance(simplified, Tagged)
    assert simplified.tags == M_ALT_GEOP.tags
    assert isinstance(simplified.reference, Scaled)
    assert simplified.reference.reference == M


def test_tagged_conversion() -> None:
    converter_ok = convert(M_ALT_GEOP, FT_ALT_GEOP)
    assert converter_ok(1) == pytest.approx(1 / 0.3048)

    with pytest.raises(DimensionMismatchError):
        _ = convert(M_ALT_GEOP, M_ALT_GEOM)
    with pytest.raises(DimensionMismatchError):
        _ = convert(M_ALT_GEOP, M)
    with pytest.raises(DimensionMismatchError):
        _ = convert(M, M_ALT_GEOP)


def test_qty_kind_call() -> None:
    from isqx import UnitKindMismatchError
    from isqx.aerospace import TRUE_AIRSPEED
    from isqx.usc import KNOT

    tas_mps = TRUE_AIRSPEED(M_PERS)
    assert isinstance(tas_mps, Tagged)
    assert tas_mps.reference == M_PERS
    assert tas_mps.tags == ("airspeed", "true")

    tas_knots = TRUE_AIRSPEED(KNOT)
    assert isinstance(tas_knots, Tagged)
    assert tas_knots.reference == KNOT
    assert tas_knots.tags == ("airspeed", "true")

    mps_to_knots = convert(tas_mps, tas_knots)
    assert mps_to_knots(1.0) == pytest.approx(1.94384449)

    with pytest.raises(UnitKindMismatchError, match="kind: `L · T⁻¹`"):
        _ = TRUE_AIRSPEED(M)

    alt_m = GEOPOTENTIAL_ALTITUDE(M)
    alt_ft = GEOPOTENTIAL_ALTITUDE(FT)
    assert convert(alt_m, alt_ft)(100) == pytest.approx(328.08399)

    with pytest.raises(DimensionMismatchError):
        _ = convert(tas_mps, alt_m)


#
# alias
#


def test_alias_fail() -> None:
    from isqx import Aliased

    with pytest.raises(CompositionError, match="can only wrap"):
        _ = Aliased(M, "fail")  # type: ignore


#
# Unit conversions
#


def test_convert_dimensionless() -> None:
    from isqx import SR

    assert convert(RAD, RAD)(1) == 1  # -> Dimensionless

    with pytest.raises(DimensionMismatchError):
        _fn = convert(RAD, SR)  # incompatible dim


def test_convert_base_dimension() -> None:
    from isqx import DIM_LENGTH, DIM_TIME

    with pytest.raises(DimensionMismatchError):
        _fn = convert(DIM_LENGTH, DIM_TIME)  # incompatible dim
    with pytest.raises(KindMismatchError):
        _fn = convert(DIM_LENGTH, dimension(RAD))  # -> Dimensionless
    assert convert(DIM_TIME, DIM_TIME)(1) == 1  # -> BaseDimension
    with pytest.raises(KindMismatchError):
        _fn = convert(1 * DIM_TIME, S)  # -> BaseUnit
    assert convert(DIM_TIME, DIM_TIME**1)(1) == 1  # -> Exp
    assert convert(DIM_TIME, Mul((DIM_TIME,)))(1) == 1  # -> Mul
    assert convert(DIM_TIME, 2 * DIM_TIME)(1) == 0.5  # -> Scaled


def test_convert_base_unit() -> None:
    with pytest.raises(DimensionMismatchError):
        _fn = convert(S, M)  # incompatible dim
    with pytest.raises(KindMismatchError):
        _fn = convert(S, RAD)  # -> Dimensionless
    assert convert(S, S)(1) == 1  # -> BaseUnit
    with pytest.raises(KindMismatchError):
        _fn = convert(S, dimension(S))  # -> BaseDimension
    assert convert(S, S**1)(1) == 1  # -> Exp
    assert convert(S, Mul((S,)))(1) == 1  # -> Mul
    assert convert(S, 2 * S)(1) == 0.5  # -> Scaled


def test_convert_exp() -> None:
    M2 = M**2
    with pytest.raises(DimensionMismatchError):
        _fn = convert(M2, M)  # incompatible dim
    with pytest.raises(KindMismatchError):
        _fn = convert(M2, RAD)  # -> Dimensionless
    with pytest.raises(KindMismatchError):
        _fn = convert(M2, dimension(M2))  # unit -> dimension
    assert convert(M2, M2)(1) == 1  # -> Exp
    assert convert(M2, Mul((M2,)))(1) == 1  # -> Mul
    assert convert(M2, 2 * M2)(1) == 0.5  # -> Scaled


def test_convert_mul() -> None:
    with pytest.raises(DimensionMismatchError):
        _fn = convert(M_PERS, M)  # incompatible dim
    with pytest.raises(KindMismatchError):
        _fn = convert(M_PERS, RAD)  # -> Dimensionless
    assert convert(M_PERS, M_PERS)(1) == 1  # -> Mul
    assert convert(M_PERS, 2 * M_PERS)(1) == 0.5  # -> Scaled

    M2_PERS2 = M_PERS**2  # would be simplified to Mul((Exp(...), ...))
    FT2_PERMIN2 = FT_PERMIN**2
    assert convert(M2_PERS2, FT2_PERMIN2)(1) == 60**2 * 0.3048**-2
    assert convert(M2_PERS2, M2_PERS2)(1) == 1


def test_convert_scaled() -> None:
    from isqx import MIN

    with pytest.raises(DimensionMismatchError):
        _fn = convert(DAY, M)  # incompatible dim
    with pytest.raises(KindMismatchError):
        _fn = convert(DAY, RAD)  # -> Dimensionless
    assert convert(DAY, S)(1) == 86400  # -> BaseUnit
    with pytest.raises(KindMismatchError):
        _value = convert(HOUR, dimension(S))  # -> BaseDimension
    assert convert(DAY, S**1)(1) == 86400  # -> Exp
    assert convert(DAY, Mul((S,)))(1) == 86400  # -> Mul
    assert convert(MIN, HOUR)(60) == 1  # -> Scaled
    assert convert(DAY, DAY)(1) == 1


def test_translated_is_terminal() -> None:
    from isqx import KILO, Translated

    with pytest.raises(CompositionError, match="exponentiated"):
        _ = CELSIUS**2
    with pytest.raises(CompositionError, match="part of a product"):
        _ = CELSIUS * M
    with pytest.raises(CompositionError, match="cannot be scaled"):
        _ = CELSIUS * 2
    with pytest.raises(CompositionError, match="prefixes cannot be applied"):
        _ = KILO * CELSIUS
    with pytest.raises(CompositionError, match="nesting"):
        _ = Translated(CELSIUS, 1, "celsius + 1")  # type: ignore
    with pytest.raises(CompositionError, match="must have a"):
        _ = Translated(M_PERS, 1, "m/s + 1")  # type: ignore


def test_convert_translated() -> None:
    from isqx import CELSIUS, DIM_TEMPERATURE, K, NonAffineConverter
    from isqx.usc import FAHRENHEIT, R

    assert dimension(CELSIUS) == DIM_TEMPERATURE

    assert convert(K, CELSIUS)(1.1) == -272.04999999999995  # inexact
    assert convert(K, CELSIUS, exact=True)(Fraction(11, 10)) == Fraction(
        -27205, 100
    )

    c_to_f = convert(CELSIUS, FAHRENHEIT, exact=True)
    assert isinstance(c_to_f, NonAffineConverter)
    assert isinstance(c_to_f.scale, Fraction) and c_to_f.scale == Fraction(9, 5)
    assert isinstance(c_to_f.offset, Fraction) and c_to_f.offset == Fraction(
        32, 1
    )
    assert c_to_f(100) == 212
    assert convert(CELSIUS, FAHRENHEIT).scale == 1.7999999999999998  # inexact

    f_to_c = convert(FAHRENHEIT, CELSIUS, exact=True)
    assert f_to_c(32) == 0
    assert f_to_c(212) == 100
    c_to_r = convert(CELSIUS, R, exact=True)
    assert c_to_r(0) == Fraction("273.15") * Fraction(9, 5)


def test_convert_tagged_translated() -> None:
    from isqx import CELSIUS, K

    SURFACE_TEMP_C = CELSIUS["surface"]
    SURFACE_TEMP_K = K["surface"]

    c_to_k_exact = convert(SURFACE_TEMP_C, SURFACE_TEMP_K, exact=True)
    assert c_to_k_exact(10) == Fraction(28315, 100)
    k_to_c_exact = convert(SURFACE_TEMP_K, SURFACE_TEMP_C, exact=True)
    assert k_to_c_exact(c_to_k_exact(10)) == 10

    with pytest.raises(DimensionMismatchError):
        convert(SURFACE_TEMP_C, K)
    with pytest.raises(DimensionMismatchError):
        convert(K, SURFACE_TEMP_C)


def test_log_level_invalid() -> None:
    from isqx import BEL, DB, DBV, HZ, KILO, Log, M, Quantity, V
    from isqx._core import _RATIO, _RatioBetween

    # NOTE: we allow logarithmic units to be composed with others,
    # TODO: in the future harden `convert` so we dont mess it up
    assert isinstance(DB**2, Exp)
    assert isinstance(DB * HZ**-1, Mul)
    assert isinstance(DB * 2, Scaled)
    assert isinstance(KILO * BEL, Scaled)

    assert isinstance(DBV**2, Exp)
    assert isinstance(DBV * M, Mul)
    assert isinstance(DBV * 2, Scaled)
    assert isinstance(KILO * DBV, Scaled)

    # other invalid compositions
    rb_tag = _RatioBetween(V, Quantity(1, V))
    with pytest.raises(CompositionError, match="only be applied to a ratio"):
        _ = M[rb_tag]
    with pytest.raises(CompositionError, match="wrap a `Dimensionless`"):
        _ = Log(DBV, base=10)  # type: ignore
    # with pytest.raises(CompositionError, match="must be physical"):
    #     _ = Relative(dimension(M), V)
    with pytest.raises(
        CompositionError, match="cannot be applied multiple times"
    ):
        _ = _RATIO[rb_tag, rb_tag]


def test_convert_logarithmic() -> None:
    from isqx import (
        DBM,
        DBUV,
        DBV,
        DBW,
        NPV,
        NPW,
        Converter,
        NonAffineConverter,
    )

    assert isinstance(dimension(DBM), Dimensionless)

    # power -> power
    dbw_to_dbm = convert(DBW, DBM)
    assert isinstance(dbw_to_dbm, NonAffineConverter)
    assert dbw_to_dbm.scale == 1
    assert dbw_to_dbm.offset == pytest.approx(30)
    assert dbw_to_dbm(10) == 40

    # root-power -> root-power
    dbv_to_dbuv = convert(DBV, DBUV, exact=True)
    assert isinstance(dbv_to_dbuv, NonAffineConverter)
    assert dbv_to_dbuv.scale == 1
    assert dbv_to_dbuv.offset == 120
    assert dbv_to_dbuv(1) == 121

    # neper <-> decibel (root-power, power)
    npv_to_dbv = convert(NPV, DBV)
    assert isinstance(npv_to_dbv, Converter)
    assert npv_to_dbv.scale == pytest.approx(20 / math.log(10))
    assert npv_to_dbv(1) == pytest.approx(8.6858896)
    dbv_to_npv = convert(DBV, NPV)
    assert isinstance(dbv_to_npv, Converter)
    assert dbv_to_npv.scale == pytest.approx(math.log(10) / 20)
    assert dbv_to_npv(20) == pytest.approx(2.302585)
    npw_to_dbw = convert(NPW, DBW)
    assert isinstance(npw_to_dbw, Converter)
    assert npw_to_dbw.scale == pytest.approx(20 / math.log(10))
    assert npw_to_dbw(1) == pytest.approx(8.6858896)


def test_convert_logarithmic_with_prefix() -> None:
    from isqx import DBV, DECI, MILLI, NPV, Converter

    npv_to_decinpv = convert(NPV, DECI * NPV, exact=True)
    assert isinstance(npv_to_decinpv, Converter)  # no offset
    assert npv_to_decinpv.scale == 10
    millinpv_to_decinpv = convert(MILLI * NPV, DECI * NPV, exact=True)
    assert millinpv_to_decinpv.scale == Fraction(1, 100)
    decinpv_to_dbv = convert(DECI * NPV, DBV)
    assert isinstance(decinpv_to_dbv, Converter)
    assert decinpv_to_dbv.scale == pytest.approx(200 / math.log(10))


def test_convert_logarithmic_fail() -> None:
    from isqx import DBM, DBV, NonLinearConversionError, V

    with pytest.raises(NonLinearConversionError):
        convert(V, DBV)
    with pytest.raises(NonLinearConversionError):
        convert(DBV, V)
    with pytest.raises(DimensionMismatchError):
        convert(DBV, DBM)


def test_angle_conversion() -> None:
    from decimal import Decimal, localcontext

    from isqx import DEG, PI, REV, E

    with localcontext() as ctx:
        assert PI.to_decimal(ctx) == Decimal("3.141592653589793238462643383")
        assert E.to_decimal(ctx) == Decimal("2.718281828459045235360287471")

    deg_to_rad = convert(DEG, RAD)
    assert deg_to_rad(180) == pytest.approx(math.pi)
    assert deg_to_rad(360) == pytest.approx(2 * math.pi)

    rad_to_deg = convert(RAD, DEG)
    assert rad_to_deg(math.pi) == pytest.approx(180.0)
    assert rad_to_deg(1) == pytest.approx(180 / math.pi)

    with localcontext() as ctx:
        ctx.prec = 100
        pi_100 = PI.to_decimal(ctx)

        assert convert(DEG, RAD, exact=True, ctx=ctx)(180) == pi_100
        assert convert(RAD, DEG, exact=True, ctx=ctx)(Fraction(pi_100)) == 180

    assert convert(REV, RAD)(1) == pytest.approx(2 * math.pi)
    assert convert(RAD, REV)(math.pi) == pytest.approx(0.5)
    assert convert(DEG, REV)(360) == pytest.approx(1.0)


def test_derived_angle_conversion() -> None:
    from isqx import DEG

    DEG_PER_S = DEG * S**-1
    RAD_PER_S = RAD * S**-1

    assert convert(DEG_PER_S, RAD_PER_S)(180) == pytest.approx(math.pi)

    result = convert(DEG_PER_S, RAD_PER_S, exact=True)(Fraction(180))
    assert isinstance(result, Fraction)
    assert result == pytest.approx(Fraction(math.pi))


#
# integration test
#


def test_xkcd_whatif_11() -> None:  # https://what-if.xkcd.com/11/
    from math import pi

    from isqx import CENTI, KILO, YEAR, BaseDimension, BaseUnit

    BIRD = BaseUnit(BaseDimension("bird"), "bird")
    POOP = BaseUnit(BaseDimension("poop"), "poop")
    MOUTH = BaseUnit(BaseDimension("mouth"), "mouth")

    PERIOD = simplify(
        (
            (BIRD * (KILO * M) ** -2)
            * (POOP * BIRD**-1 * HOUR**-1)
            * (HOUR * DAY**-1)
            * (MOUTH * POOP**-1)
            * ((CENTI * M) ** 2 * MOUTH**-1)
        )
        ** -1
    )  # = (km^2 * day) / cm^2
    assert isinstance(PERIOD, Scaled)
    assert isinstance(PERIOD.factor, LazyProduct)
    assert PERIOD.factor.to_exact(ctx=CTX_DEFAULT) == 100_000**2 * 86400
    assert PERIOD.reference == S

    num_birds = 300e9
    earth_surface_area = 4 * pi * 6378**2
    period_yr = convert(PERIOD, YEAR)(
        1
        / (
            (num_birds / earth_surface_area)  # bird / km^2
            * 1  # poop / (bird * hour)
            * 16  # hours / day
            * 1  # mouth / poop
            * 15  # cm^2 / mouth
        )
    )
    assert period_yr == pytest.approx(195, abs=0.7)

    from isqx import MILLI, YEAR
    from isqx.usc import FL_OZ, MI, MPG

    MM2 = (MILLI * M) ** 2
    assert convert(MPG**-1, MM2)(1 / 20) == pytest.approx(0.11760729)
    poop_dropping_rate = 0.5  # fl_oz / (day * bird)
    total_distance_driven_rate = 3e12  # mi / year
    assert 1 / (
        convert(
            BIRD  #
            * (FL_OZ * DAY**-1 * BIRD**-1)
            * (MI * YEAR**-1) ** -1,
            MPG**-1,
        )(num_birds * poop_dropping_rate / total_distance_driven_rate)
    ) == pytest.approx(7.009, rel=1e-3)  #  xkcd's 13MPG is wrong



#!/usr/bin/env -S uv run --script
# /// script
# requires-python = ">=3.9"
# dependencies = [
#     "typer",
#     "httpx",
#     "polars",
# ]
# ///
from __future__ import annotations

import json
import logging
import re
from datetime import datetime
from pathlib import Path
from typing import Any, Generator, Iterable, Literal

import httpx
import polars as pl

ENDPOINT_WIKIDATA_SPARQL = "https://query.wikidata.org/sparql"
PATH_BASE = Path(__file__).parent / "wikidata"
FP_QUANTITIES_RQ = PATH_BASE / "quantities.sparql"
FP_QUANTITIES_JSON = PATH_BASE / "quantities.json"
FP_QUANTITIES_PQ = PATH_BASE / "quantities.parquet"
FP_QUANTITIES_MD = PATH_BASE / "quantities.md"

logger = logging.getLogger(__name__)


def fetch(
    query: str,
    *,
    fmt: Literal["json", "xml"],
    endpoint_url: str = ENDPOINT_WIKIDATA_SPARQL,
) -> bytes:
    response = httpx.post(
        endpoint_url, data={"query": query, "format": fmt}, timeout=120.0
    )
    response.raise_for_status()
    return response.content


def fetch_json(
    fp_rq: Path = FP_QUANTITIES_RQ,
    fp_json: Path = FP_QUANTITIES_JSON,
) -> None:
    """Execute a SPARQL query from a file and save the raw response as JSON."""
    if fp_json.exists():
        logger.info(f"file already exists, skipping fetch: {fp_json}")
        return
    query = fp_rq.read_text()
    res = fetch(query, fmt="json")
    with open(fp_json, "wb") as f:
        f.write(res)


#
# parquet
#

RE_CLAUSE = r"(?P<a>\d+)[-.](?P<b>\d+)(\.(?P<c>[\d\w]+))?"


def to_int(s: str | None) -> int | None:
    if s is None:
        return None
    try:
        return int(s)
    except ValueError:
        if len(s) != 1:
            logger.warning(f"expected char of length 1, but got {s}")
            return None
        val = ord(s.lower()) - ord("a")
        if val < 0 or val > 25:
            logger.warning(f"expected [a-z], but got {s}")
            return None
        return val


def parse_isoiec_items(
    isoiec_items: Iterable[str],
) -> Generator[dict[str, Any], None, None]:
    for isoiec_item in isoiec_items:
        if not isoiec_item:
            continue
        t_string, clause = items(isoiec_item, separator="|||")
        dt = datetime.strptime(t_string, "%Y-%m-%dT%H:%M:%SZ")
        if not (match := re.match(RE_CLAUSE, clause)):
            clause_order = None
        else:
            clause_order = (
                (to_int(match.group("a")) or 0) * 10000
                + (to_int(match.group("b")) or 0) * 100
                + (to_int(match.group("c")) or 0)
            )

        yield {"dt": dt, "clause": clause, "clause_order": clause_order}


def shorten_uri(uri: str) -> str:
    code = uri.removeprefix("http://www.wikidata.org/entity/")
    if code[0] in ("Q", "P"):
        return code
    return uri


def parse_uri(value: dict[str, str] | None) -> str | None:
    if not value:
        return None
    if (t := value["type"]) != "uri":
        raise ValueError(f"expected uri type, got {t}")
    return shorten_uri(value["value"])


def parse_literal(value: dict[str, str] | None) -> str | None:
    if not value:
        return None
    if (t := value["type"]) != "literal":
        raise ValueError(f"expected literal type, got {t}")
    return value["value"]


def items(
    lit: str | None, *, separator: str = "@@@"
) -> Generator[str, None, None]:
    if not lit:
        return
    for item in lit.split(separator):
        yield item


def parse_latex(mathml: str) -> str:
    anno = mathml.split('<annotation encoding="application/x-tex">', 1)[-1]
    anno = anno.split("</annotation>", 1)[0]
    if anno.startswith(r"{\displaystyle "):
        anno = anno.removeprefix(r"{\displaystyle ").removesuffix(r"}")
    return anno


def parse_symbol_in_formula(sym: str | None) -> dict[str, str] | None:
    if not sym:
        return None
    if len(i := tuple(items(sym, separator="|||"))) != 3:
        logger.warning(
            f"expected 3 items in symbol_in_formula, got {len(i)}: {sym}"
        )
        return None
    mathml, label, uri = i
    return {
        "latex": parse_latex(mathml),
        "label": label,
        "entity": shorten_uri(uri),
    }


def parse_response_quantities(
    bindings: Iterable[dict[str, Any]],
) -> Generator[dict[str, Any], None, None]:
    for binding in bindings:
        # one binding can be referenced by multiple ISO/IEC 80000 editions
        yield {
            "entity": parse_uri(binding.get("quantity")),
            "isoiec_items": sorted(
                parse_isoiec_items(
                    items(parse_literal(binding.get("isoiec_items")))
                ),
                key=lambda x: x["dt"],
                reverse=True,
            ),
            "label": parse_literal(binding.get("label")),
            "desc": parse_literal(binding.get("desc")),
            "symbols": tuple(
                parse_latex(mathml)
                for mathml in items(parse_literal(binding.get("symbols")))
            ),
            "dims": tuple(
                parse_latex(mathml)
                for mathml in items(parse_literal(binding.get("dims")))
            ),
            "units": tuple(
                parse_latex(mathml)
                for mathml in items(parse_literal(binding.get("units")))
            ),
            "wolfram_ids": tuple(
                items(parse_literal(binding.get("wolfram_ids")))
            ),
            "qudt_ids": tuple(items(parse_literal(binding.get("qudt_ids")))),
            "defining_formulas": tuple(
                parse_latex(mathml)
                for mathml in items(
                    parse_literal(binding.get("defining_formulas"))
                )
            ),
            "symbols_in_formula": tuple(
                parse_symbol_in_formula(sym)
                for sym in items(
                    parse_literal(binding.get("symbols_in_formula"))
                )
            ),
        }


def quantities_to_parquet(
    fp_json: Path = FP_QUANTITIES_JSON, fp_parquet: Path = FP_QUANTITIES_PQ
) -> None:
    """Organise the raw quantities JSON response in a parquet file."""
    response = json.loads(fp_json.read_text())
    bindings = response["results"]["bindings"]
    df = pl.DataFrame(parse_response_quantities(bindings))
    df.write_parquet(fp_parquet)
    logger.info(f"wrote {len(df)} rows to {fp_parquet}")


def md_lines(row: dict[str, Any]) -> Generator[str, None, None]:
    clause = ", ".join(
        f"`{item['dt'].year}::{item['clause']}`" for item in row["isoiec_items"]
    )
    clause = f"{clause} = " if clause else ""
    yield f"**{clause}`{row['entity']}`: {row['label']}**"
    if desc := row["desc"]:
        yield f'- description: "{desc}"'
    if dims := row["dims"]:
        yield f"- dimension: ${', '.join(dims)}$"
    if sym := row["symbols"]:
        yield f"- symbol: ${', '.join(sym)}$"
    if de := row["defining_formulas"]:
        yield f"- formula: ${', '.join(de)}$"
        for sym in row["symbols_in_formula"]:
            yield f"  - ${sym['latex']}$ = {sym['label']}"


def quantities_to_md(
    fp_parquet: Path = FP_QUANTITIES_PQ,
    fp_md: Path = FP_QUANTITIES_MD,
) -> None:
    """Organise the quantities parquet file in a human-readable markdown file."""
    df = (
        pl.scan_parquet(fp_parquet)
        .filter(pl.col("isoiec_items").list.len() > 0)
        .sort(pl.col("isoiec_items").list.first().struct.field("clause_order"))
        .collect()
    )
    with open(fp_md, "w") as f:
        for row in df.iter_rows(named=True):
            for line in md_lines(row):
                f.write(line + "\n")
            f.write("\n")
    logger.info(f"wrote {len(df)} entries to {fp_md}")


try:
    import typer

    app = typer.Typer(no_args_is_help=True)
    app.command()(fetch_json)
    app.command()(quantities_to_parquet)
    app.command()(quantities_to_md)
except ImportError:

    def app() -> None:
        fetch_json()
        quantities_to_parquet()
        quantities_to_md()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    app()



"""
Units and quantities common in aerospace engineering.

See: [isqx._citations.ICAO][]
"""
# TODO: ISO 2533:1975 (standard atmosphere)
# TODO: ISO 1151

from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal
from typing import Annotated, Literal

from ._core import (
    DELTA,
    Dimensionless,
    OriginAt,
    QtyKind,
    Quantity,
    ratio,
    slots,
)
from ._iso80000 import (
    ALTITUDE,
    AREA,
    DISTANCE,
    DRAG,
    DRAG_COEFFICIENT,
    DYNAMIC_PRESSURE,
    HOUR,
    KG,
    LENGTH,
    LIFT,
    M_PERS,
    MACH_NUMBER,
    MASS,
    MASS_FLOW_RATE,
    MIN,
    MOMENT_OF_FORCE,
    PA,
    POWER,
    PRESSURE,
    RAD,
    RAD_PERS,
    SPECIFIC_ENERGY,
    TEMPERATURE,
    VELOCITY,
    K,
    L,
    M,
    N,
    S,
    W,
)
from .usc import FT

#
# aircraft performance: state
#

# heading: [0, 360) degrees
HEADING = QtyKind(RAD, ("heading",))
HEADING_TRUE = HEADING["true"]
HEADING_MAG = HEADING["magnetic"]
HEADING_TRUE_WIND = HEADING_TRUE["wind"]
HEADING_MAG_WIND = HEADING_MAG["wind"]
GROUND_TRACK = HEADING["ground_track"]
"""Direction of the aircraft's velocity vector relative to the ground."""

PRESSURE_ALTITUDE = ALTITUDE["pressure"]
"""Pressure altitude, as measured by the altimeter (standard pressure setting 1013.25 hPa)."""
DENSITY_ALTITUDE = ALTITUDE["density"]
"""Density altitude, as measured by the altimeter."""
GEOPOTENTIAL_ALTITUDE = ALTITUDE["geopotential"]
"""Geopotential altitude, as measured from mean sea level."""
GEOMETRIC_ALTITUDE = ALTITUDE["geometric"]
"""Altitude measured from mean sea level (e.g. via GNSS)."""
# height: measured from *specific* datum
GEODETIC_HEIGHT = QtyKind(M, ("height", "geodetic"))
"""Height above the reference ellipsoid."""
HEIGHT_ABOVE_GROUND_LEVEL = QtyKind(M, ("height", "above_ground_level"))
"""Height above ground level (radio altimeter)."""

L_OVER_D = ratio(LIFT(N), DRAG(N))

K_PERM = K * M**-1
"""Kelvin per meter, a unit of temperature gradient. For use in ISA."""
ENERGY_HEIGHT = LENGTH["energy_height"]
"""Specific energy expressed as a height."""
SPECIFIC_EXCESS_POWER = QtyKind(M_PERS, ("specific_excess_power",))

#
# aircraft geometry
#

WINGSPAN = LENGTH["wingspan"]
CHORD = LENGTH["chord"]
MEAN_AERODYNAMIC_CHORD = CHORD["mean_aerodynamic"]
"""Mean aerodynamic chord (MAC)."""
MEAN_GEOMETRIC_CHORD = CHORD["mean_geometric"]
"""Mean geometric chord (Standard Mean Chord)."""
WING_AREA = AREA["wing"]
"""Reference wing area."""
WETTED_AREA = AREA["wetted"]
PLANFORM_AREA = AREA["planform"]
FRONTAL_AREA = AREA["frontal"]
"""Cross-sectional area perpendicular to the flow."""
DISK_AREA = AREA["disk"]
"""Area swept by a propeller or rotor."""
TAIL_AREA = AREA["tail"]
TAIL_MOMENT_ARM = LENGTH["tail_moment_arm"]
ASPECT_RATIO = Dimensionless("aspect_ratio")
TAPER_RATIO = Dimensionless("taper_ratio")
SWEEP_ANGLE = QtyKind(RAD, ("angle", "sweep"))
DIHEDRAL_ANGLE = QtyKind(RAD, ("angle", "dihedral"))
TWIST_ANGLE = QtyKind(RAD, ("angle", "twist"))
"""Washout or washin angle."""
FINENESS_RATIO = Dimensionless("fineness_ratio")
"""Ratio of length to maximum diameter for a fuselage or body."""

#
# aerodynamics
#

ANGLE_OF_ATTACK = QtyKind(RAD, ("angle", "angle_of_attack"))
"""Angle between the chord line and the relative wind vector."""
SIDESLIP_ANGLE = QtyKind(RAD, ("angle", "sideslip"))
"""Angle between the relative wind vector and the plane of symmetry."""
DOWNWASH_ANGLE = QtyKind(RAD, ("angle", "downwash"))

CRITICAL_MACH_NUMBER = MACH_NUMBER["critical"]
DRAG_DIVERGENCE_MACH_NUMBER = MACH_NUMBER["drag_divergence"]

ZERO_LIFT_DRAG_COEFFICIENT = DRAG_COEFFICIENT["zero_lift"]
LIFT_INDUCED_DRAG_COEFFICIENT = DRAG_COEFFICIENT["lift_induced"]
INDUCED_DRAG_COEFFICIENT = DRAG_COEFFICIENT["induced"]

OSWALD_EFFICIENCY = Dimensionless("oswald_efficiency_factor")
"""Span efficiency factor."""
# TODO: stability derivatives
PITCHING_MOMENT_COEFFICIENT = Dimensionless("pitching_moment_coefficient")
ROLLING_MOMENT_COEFFICIENT = Dimensionless("rolling_moment_coefficient")
YAWING_MOMENT_COEFFICIENT = Dimensionless("yawing_moment_coefficient")
PRESSURE_COEFFICIENT = Dimensionless("pressure_coefficient")
SKIN_FRICTION_COEFFICIENT = Dimensionless("skin_friction_coefficient")
LIFT_SLOPE = Dimensionless("lift_slope")
"""Change in lift coefficient per unit angle of attack (per radian)."""

CIRCULATION = QtyKind(M**2 * S**-1, ("circulation",))

#
# aircraft design
#

AIRCRAFT_MASS = MASS["aircraft"]
GROSS = AIRCRAFT_MASS["gross"]
CARGO_CAPACITY = AIRCRAFT_MASS["cargo_capacity"]
FUEL_CAPACITY = AIRCRAFT_MASS["fuel_capacity"]
TAKEOFF_MASS = AIRCRAFT_MASS["takeoff"]
LANDING_MASS = AIRCRAFT_MASS["landing"]
MAXIMUM_TAKEOFF_WEIGHT = TAKEOFF_MASS["maximum"]
ZERO_FUEL_WEIGHT = AIRCRAFT_MASS["zero_fuel_weight"]
OPERATING_EMPTY_WEIGHT = AIRCRAFT_MASS["operating_empty"]
PAYLOAD = AIRCRAFT_MASS["payload"]
EMPTY_WEIGHT = AIRCRAFT_MASS["empty"]

TANK_CAPACITY = QtyKind(L, ("aircraft", "tank_capacity"))  # ICAO 1.14
ENDURANCE = QtyKind(HOUR, ("aircraft", "endurance"))  # ICAO 1.6

WING_LOADING = QtyKind(N * M**-2, ("wing_loading",))
"""Weight of the aircraft divided by the wing area."""
POWER_LOADING = QtyKind(N * W**-1, ("power_loading",))
"""Weight of the aircraft divided by the engine power."""
THRUST_LOADING = Dimensionless("thrust_loading")
"""Thrust to weight ratio."""

#
# flight dynamics
#

LOAD_FACTOR = Dimensionless("load_factor")
"""Ratio of lift to weight (n)."""

ANGULAR_VELOCITY = QtyKind(RAD_PERS, ("angular_velocity",))
ROLL_RATE = ANGULAR_VELOCITY["roll"]
"""Angular velocity about the body X axis."""
PITCH_RATE = ANGULAR_VELOCITY["pitch"]
"""Angular velocity about the body Y axis."""
YAW_RATE = ANGULAR_VELOCITY["yaw"]
"""Angular velocity about the body Z axis."""
TURN_RATE = ANGULAR_VELOCITY["turn"]
"""Rate of change of heading."""

ATTITUDE = QtyKind(RAD, ("attitude",))
BANK_ANGLE = ATTITUDE["bank"]
PITCH_ANGLE = ATTITUDE["pitch"]
FLIGHT_PATH_ANGLE = ATTITUDE["flight_path"]
"""Angle between the velocity vector and the horizon."""

AIRCRAFT_MOMENT = MOMENT_OF_FORCE["aircraft"]
PITCHING_MOMENT = AIRCRAFT_MOMENT["pitching"]
ROLLING_MOMENT = AIRCRAFT_MOMENT["rolling"]
YAWING_MOMENT = AIRCRAFT_MOMENT["yawing"]

#
# stability and control
#

STATIC_MARGIN = Dimensionless("static_margin")
"""Distance between the neutral point and the center of gravity, normalized by MAC."""
NEUTRAL_POINT = DISTANCE["neutral_point"]
"""Longitudinal position of the aerodynamic center of the whole aircraft."""
CENTER_OF_GRAVITY = DISTANCE["center_of_gravity"]
TAIL_VOLUME_COEFFICIENT = Dimensionless("tail_volume_coefficient")
HORIZONTAL_TAIL_VOLUME_COEFFICIENT = TAIL_VOLUME_COEFFICIENT["horizontal"]
VERTICAL_TAIL_VOLUME_COEFFICIENT = TAIL_VOLUME_COEFFICIENT["vertical"]

#
# aircraft performance
#

STATIC_TEMPERATURE = TEMPERATURE["static"]
TOTAL_TEMPERATURE = TEMPERATURE["total"]
"""Also known as stagnation temperature."""
CONST_TEMPERATURE_ISA: Annotated[Decimal, STATIC_TEMPERATURE(K)] = Decimal(
    "288.15"
)
TEMPERATURE_DEVIATION_ISA = STATIC_TEMPERATURE[
    DELTA, OriginAt(Quantity(CONST_TEMPERATURE_ISA, K))
]
"""Deviation from the [ISA temperature at sea level][isqx.aerospace.CONST_TEMPERATURE_ISA]."""

TOTAL_PRESSURE = PRESSURE["total"]
IMPACT_PRESSURE = DYNAMIC_PRESSURE["impact"]

PRESSURE_RATIO = Dimensionless("pressure_ratio")
"""Ratio of static pressure to standard sea level pressure."""
TEMPERATURE_RATIO = Dimensionless("temperature_ratio")
"""Ratio of static temperature to standard sea level temperature."""
DENSITY_RATIO = Dimensionless("density_ratio")
"""Ratio of air density to standard sea level density."""

# linear velocity
AIRSPEED = QtyKind(M_PERS, ("airspeed",))
INDICATED_AIRSPEED = AIRSPEED["indicated"]
"""Indicated airspeed (IAS), as measured directly from the pitot-static system."""
CALIBRATED_AIRSPEED = AIRSPEED["calibrated"]
"""Calibrated airspeed (CAS), [IAS][isqx.aerospace.INDICATED_AIRSPEED] corrected for instrument and position errors."""
EQUIVALENT_AIRSPEED = AIRSPEED["equivalent"]
"""Equivalent airspeed (EAS), [CAS][isqx.aerospace.CALIBRATED_AIRSPEED] corrected for compressibility."""
TRUE_AIRSPEED = AIRSPEED["true"]
"""True airspeed (TAS), speed relative to the airmass."""
GROUND_SPEED = AIRSPEED["ground"]
"""Speed relative to the ground."""
STALL_SPEED = AIRSPEED["stall"]
APPROACH_SPEED = AIRSPEED["approach"]
TAKEOFF_SPEED = AIRSPEED["takeoff"]
ROTATE_SPEED = AIRSPEED["rotate"]
V1_SPEED = AIRSPEED["v1"]
V2_SPEED = AIRSPEED["v2"]
VREF_SPEED = AIRSPEED["vref"]
# TODO: other v speeds
CORNER_SPEED = AIRSPEED["corner"]
"""The speed at which the maximum lift coefficient and the maximum load factor are reached simultaneously."""
WIND_SPEED = QtyKind(M_PERS, ("wind",))
"""Wind speed."""
SPEED_OF_SOUND = QtyKind(M_PERS, ("sound",))
"""Speed of sound."""

FT_PER_MIN = FT * MIN**-1
VERTICAL_RATE = QtyKind(M_PERS, ("vertical_rate",))
"""Rate of climb or descent.

Commonly expressed in [feet per minute][isqx.aerospace.FT_PER_MIN]."""
VERTICAL_RATE_INERTIAL = VERTICAL_RATE["inertial"]
"""Vertical rate derived from inertial sensors/GNSS."""
VERTICAL_RATE_BAROMETRIC = VERTICAL_RATE["barometric"]
"""Vertical rate derived from barometric pressure changes."""

SPECIFIC_IMPULSE = QtyKind(S, ("specific_impulse",))
RANGE = DISTANCE["range"]
TAKEOFF_DISTANCE = DISTANCE["takeoff"]
LANDING_DISTANCE = DISTANCE["landing"]
TURN_RADIUS = DISTANCE["turn_radius"]

#
# propulsion
#
SHAFT_POWER = POWER["shaft"]
"""Power delivered to a shaft (e.g. turboprop)."""
BRAKE_POWER = POWER["brake"]
EQUIVALENT_SHAFT_POWER = SHAFT_POWER["equivalent"]

ENGINE_MASS_FLOW_RATE = MASS_FLOW_RATE["engine"]
KG_PERS = KG * S**-1
THRUST_SPECIFIC_FUEL_CONSUMPTION = QtyKind(KG_PERS * N**-1, ("engine",))
"""Fuel mass flow rate per unit thrust."""
POWER_SPECIFIC_FUEL_CONSUMPTION = QtyKind(
    KG_PERS * W**-1, ("engine", "power_specific")
)
"""Fuel mass flow rate per unit power."""
FUEL_SPECIFIC_ENERGY = SPECIFIC_ENERGY["fuel"]
EXHAUST_VELOCITY = VELOCITY["exhaust"]
BYPASS_RATIO = Dimensionless("bypass_ratio")
# TODO: make efficieny kinds more specific
PROPULSIVE_EFFICIENCY = Dimensionless("efficiency_propulsive")
PROPELLER_EFFICIENCY = Dimensionless("efficiency_propeller")
ADVANCE_RATIO = Dimensionless("advance_ratio")
"""Ratio of freestream speed to tip speed for propellers."""

#
# aeroacoustics
#
# TODO: dBA, EPNdB etc.

#
# navigation
#


@dataclass(frozen=True, **slots)
class Aerodrome:
    ident: str
    ident_kind: Literal["icao", "iata"] | str


PRESSURE_ALTIMETER = QtyKind(PA, ("altimeter",))
"""Altimeter setting (QNH/QFE)."""
RUNWAY_LENGTH = QtyKind(M, ("runway", "length"))  # ICAO 1.12
RUNWAY_VISUAL_RANGE = QtyKind(M, ("runway", "visual_range"))  # ICAO 1.13
VISIBILITY = QtyKind(M, ("meteo", "visibility"))  # ICAO 1.15

#
# adsb/mode s
#

ICAO_ADDRESS = Dimensionless("icao_address_24_bit")
"""Unique 24-bit aircraft address assigned by ICAO."""
SQUAWK_CODE = Dimensionless("squawk_code_12_bit")
"""Mode A code (4 octal digits)."""

NAVIGATION_UNCERTAINTY_CATEGORY_POSITION = Dimensionless("adsb_nucp")
NAVIGATION_UNCERTAINTY_CATEGORY_VELOCITY = Dimensionless("adsb_nucv")
NAVIGATION_ACCURACY_CATEGORY_POSITION = Dimensionless("adsb_nacp")
NAVIGATION_ACCURACY_CATEGORY_VELOCITY = Dimensionless("adsb_nacv")
NAVIGATION_INTEGRITY_CATEGORY = Dimensionless("adsb_nic")
SURVEILLANCE_INTEGRITY_LEVEL = Dimensionless("adsb_sil")



"""A `MkDocs` plugin bundling a default `KaTeX` configuration necessary to
render mathematical expressions in the documentation.

[`isqx.mkdocs.extension.IsqxExtension`][] is responsible for injecting
[cross-referenced details][isqx.details.Details] into the docstrings of
attributes and functions. This module also comes with an optional client-side JS
script `detail-highlight.js` that searches for matching symbols within
[`isqx.details.Detail`][] blocks and highlights them.

!!! note

    You must install `isqx` with the `docs` extra optional dependencies
    to use this module.
"""

from __future__ import annotations

from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, Generator

from mkdocs.config import Config
from mkdocs.config.config_options import ExtraScriptValue, Type
from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.utils import copy_file
from mkdocstrings import MkdocstringsPlugin

from . import PATH_PLUGIN_ASSETS

if TYPE_CHECKING:
    from mkdocs.config.defaults import MkDocsConfig

logger = getLogger(__name__)


class IsqxPluginConfig(Config):  # type: ignore
    definitions = Type(list, default=[])
    """Modules that contain the definitions, e.g. `isqx.iso80000`, `isqx.aerospace`.

    This is used to provide cross-references within the [`isqx.details.Detail`][]
    blocks.

    Note that `griffe` will **dynamically import** these modules.
    To reduce build time, avoid expensive imports like `torch`."""
    details = Type(list, default=[])
    """Paths to the details mapping, e.g. `isqx.details.iso80000.TIME_AND_SPACE`."""
    details_highlight = Type(bool, default=True)
    """Whether to highlight symbols in the details."""
    load_katex = Type(bool, default=True)
    """Whether to load KaTeX javascript and CSS assets."""


PLUGIN_ASSETS_JS = (PATH_PLUGIN_ASSETS / "js" / "detail-highlight.js",)
KATEX_VERSION = "0.16.25"


class IsqxPlugin(BasePlugin[IsqxPluginConfig]):  # type: ignore
    def on_config(self, config: MkDocsConfig) -> MkDocsConfig:
        plugin = config.plugins.get("mkdocstrings")
        assert isinstance(plugin, MkdocstringsPlugin)
        options = (
            plugin.config.setdefault("handlers", {})
            .setdefault("python", {})
            .setdefault("options", [])
        )
        extensions = options.setdefault("extensions", [])
        extensions.append(
            {
                "isqx.mkdocs.extension:IsqxExtension": {
                    "config": self.config,
                    "objects_out_path": f"{config.site_dir}/assets",
                }
            }
        )

        self.extra_css: list[str] = []
        self.extra_scripts: list[ExtraScriptValue | str] = []
        if self.config.details_highlight:
            self.extra_scripts.append(_mjs("js/cmap.mjs"))
            self.extra_scripts.append(_mjs("js/detail-highlight.mjs"))
            self.extra_css.append("css/detail-highlight.css")
        if self.config.load_katex:
            self.extra_scripts.append(_mjs("js/katex.mjs"))
            self.extra_scripts.extend(katex_js(KATEX_VERSION))
            self.extra_css.extend(katex_css(KATEX_VERSION))
        if self.extra_scripts:
            config.extra_javascript[:0] = self.extra_scripts
        if self.extra_css:
            config.extra_css[:0] = self.extra_css
        return config

    def on_post_build(self, *, config: MkDocsConfig) -> None:
        output_path = Path(config.site_dir)
        for script in self.extra_scripts:
            maybe_copy_asset(script, str(output_path))
        for css in self.extra_css:
            maybe_copy_asset(css, str(output_path))


def katex_js(version: str) -> Generator[ExtraScriptValue, None, None]:
    yield _script(
        f"https://cdn.jsdelivr.net/npm/katex@{version}/dist/katex.min.js"
    )
    yield _script(
        f"https://cdn.jsdelivr.net/npm/katex@{version}/dist/contrib/auto-render.min.js"
    )
    yield _script(
        f"https://cdn.jsdelivr.net/npm/katex@{version}/dist/contrib/copy-tex.min.js"
    )


def _script(path: str) -> ExtraScriptValue:
    s: ExtraScriptValue = ExtraScriptValue(path)
    s.defer = True
    return s


def _mjs(path: str) -> ExtraScriptValue:
    s: ExtraScriptValue = ExtraScriptValue(path)
    s.type = "module"
    s.defer = True
    return s


def katex_css(version: str) -> Generator[str, None, None]:
    yield f"https://cdn.jsdelivr.net/npm/katex@{version}/dist/katex.min.css"


def maybe_copy_asset(asset: ExtraScriptValue | str, output_path: str) -> None:
    path = asset.path if isinstance(asset, ExtraScriptValue) else asset
    if path.startswith("js/") or path.startswith("css/"):
        a = PATH_PLUGIN_ASSETS / path
        copy_file(
            str(a),
            str(Path(output_path) / a.relative_to(PATH_PLUGIN_ASSETS)),
        )



"""Citations for `isqx`.

They can be used for cross-references in the documentation, e.g.
`see [special publication 811][isqx._citations.SP811]`.
A tiny griffe extension is used to copy the runtime value into the docstring.
"""

SI = r"""```bibtex
@techreport{si,
    author = {{Bureau International des Poids et Mesures}},
    title = {The International System of Units ({SI})},
    edition = {9th},
    year = {2019},
    note = {Version 3.01, updated August 2024},
    institution = {Bureau International des Poids et Mesures (BIPM)},
    address = {Sèvres, France},
    url = {https://www.bipm.org/documents/20126/41483022/SI-Brochure-9-EN.pdf}
}
```"""
IUPAP1 = r"""```bibtex
@techreport{iupap1,
    author = {{IUPAP Commission C2 on Symbols, Units, Nomenclature, Atomic Masses and Fundamental Constants}},
    title = {Symbols, Units, Nomenclature and Fundamental Constants in Physics},
    year = {1987},
    institution = {International Union of Pure and Applied Physics (IUPAP)},
    url = {https://archive2.iupap.org/wp-content/uploads/2014/05/A4.pdf}
}
```"""
SP811 = r"""```bibtex
@techreport{sp811,
    title = {Guide for the Use of the International System of Units ({SI})},
    year = {2008},
    edition = {2008},
    institution = {National Institute of Standards and Technology},
    type = {Special Publication},
    number = {811},
    address = {Gaithersburg, MD},
    url = {https://www.nist.gov/pml/special-publication-811}
}
```"""
CODATA2022 = r"""```bibtex
@techreport{codata2022,
    title = {{CODATA} Recommended Values of the Fundamental Physical Constants: 2022},
    year = {2024},
    month = {May},
    institution = {National Institute of Standards and Technology},
    type = {Special Publication},
    number = {961},
    url = {https://physics.nist.gov/cuu/pdf/sp961.pdf}
}
```"""
ISO_80000_2 = r"""```bibtex
@standard{iso_80000_2,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 2: Mathematics},
    year = {2019},
    number = {ISO 80000-2:2019},
    url = {https://www.iso.org/standard/76921.html}
}
```"""
ISO_80000_3 = r"""```bibtex
@standard{iso_80000_3,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 3: Space and time},
    year = {2019},
    number = {ISO 80000-3:2019},
    url = {https://www.iso.org/standard/64974.html}
}
```"""
ISO_80000_4 = r"""```bibtex
@standard{iso_80000_4,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 4: Mechanics},
    year = {2019},
    number = {ISO 80000-4:2019},
    url = {https://www.iso.org/standard/64975.html}
}
```"""
ISO_80000_5 = r"""```bibtex
@standard{iso_80000_5,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 5: Thermodynamics},
    year = {2019},
    number = {ISO 80000-5:2019},
    url = {https://www.iso.org/standard/64976.html}
}
```"""
IEC_80000_6 = r"""```bibtex
@standard{iec_80000_6,
    organization = {International Electrotechnical Commission},
    title = {Quantities and units - Part 6: Electromagnetism},
    year = {2022},
    number = {IEC 80000-6:2022},
    url = {https://www.iso.org/standard/77846.html}
}
```"""
IEV = r"""```bibtex
@misc{iev,
    author = {{IEC}},
    title = {Electropedia - International Electrotechnical Vocabulary},
    organization = {International Electrotechnical Commission},
    url = {https://www.electropedia.org/}
}
```"""
ISO_80000_7 = r"""```bibtex
@standard{iso_80000_7,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 7: Light and radiation},
    year = {2019},
    number = {ISO 80000-7:2019},
    url = {https://www.iso.org/standard/64977.html}
}
```"""
ISO_80000_8 = r"""```bibtex
@standard{iso_80000_8,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 8: Acoustics},
    year = {2020},
    number = {ISO 80000-8:2020},
    url = {https://www.iso.org/standard/64978.html}
}
```"""
ISO_80000_9 = r"""```bibtex
@standard{iso_80000_9,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 9: Physical chemistry and molecular physics},
    year = {2019},
    number = {ISO 80000-9:2019},
    url = {https://www.iso.org/standard/64979.html}
}
```"""
ISO_80000_10 = r"""```bibtex
@standard{iso_80000_10,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 10: Atomic and nuclear physics},
    year = {2019},
    number = {ISO 80000-10:2019},
    url = {https://www.iso.org/standard/64980.html}
}
```"""
ISO_80000_11 = r"""```bibtex
@standard{iso_80000_11,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 11: Characteristic numbers},
    year = {2019},
    number = {ISO 80000-11:2019},
    url = {https://www.iso.org/standard/64982.html}
}
```"""
ISO_80000_12 = r"""```bibtex
@standard{iso_80000_12,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 12: Solid state physics},
    year = {2019},
    number = {ISO 80000-12:2019},
    url = {https://www.iso.org/standard/63480.html}
}
```"""
ISO_80000_13 = r"""```bibtex
@standard{iso_80000_13,
    organization = {International Organization for Standardization},
    title = {Quantities and units - Part 13: Information science and technology},
    year = {2025},
    number = {ISO 80000-13},
    url = {https://www.iso.org/standard/87648.html}
}
```"""
MP_UNITS = r"""```bibtex
@software{mp_units,
    author = {Pusz, Mateusz and Guerrero Peña, Johel Ernesto and Hogg, Chip and {The mp-units project team}},
    license = {MIT},
    month = nov,
    title = {{mp-units}},
    url = {https://github.com/mpusz/mp-units},
    version = {2.4.0},
    year = {2024}
}
```"""
# usc
H44 = r"""```bibtex
@techreport{h44,
    author = {{NIST}},
    title = {NIST Handbook 44 - 2024 - Appendix C. General Tables of Units of Measurement},
    year = {2024},
    institution = {National Institute of Standards and Technology},
    url = {https://www.nist.gov/document/nist-handbook-44-2024-appendix-c-pdf}
}
```"""
WMA1985 = r"""```bibtex
@misc{wma1985,
    title = {Weights and Measures Act 1985},
    year = {1985},
    organization = {{legislation.gov.uk}},
    url = {https://www.legislation.gov.uk/ukpga/1985/72/contents}
}
```"""
# aerospace
ICAO = r"""```bibtex
@misc{icao,
    author = {{ICAO}},
    title = {Annex 5 - Units of Measurement to be Used in the Air and Ground Services},
    organization = {International Civil Aviation Organization},
    url = {https://store.icao.int/en/annex-5-units-of-measurement-to-be-used-in-the-air-and-ground-services}
}
```"""



from __future__ import annotations

from dataclasses import dataclass
from decimal import (
    Decimal,
    InvalidOperation,
    getcontext,
    localcontext,
)
from fractions import Fraction

import pytest
from isqx import (
    LazyProduct,
    Number,
)

LP = 8
DP = 28
HP = 50


@dataclass
class LFCase:
    products_input: tuple[tuple[Number, Fraction | int] | Number, ...]
    expected_value: Fraction | str | None
    precision: int | None
    expected_exception: type[Exception] | None


LF_CASES = [
    # no exponent
    LFCase(tuple(), Fraction(1), DP, None),
    LFCase((0,), Fraction(0), DP, None),
    LFCase((1,), Fraction(1), DP, None),
    LFCase((5,), Fraction(5), DP, None),
    LFCase((-5,), Fraction(-5), DP, None),
    LFCase((Fraction(3, 4),), Fraction(3, 4), DP, None),
    LFCase((2.5,), Fraction(5, 2), DP, None),
    LFCase((0.125,), Fraction(1, 8), DP, None),
    LFCase((0.1,), Fraction(3602879701896397, 36028797018963968), DP, None),
    LFCase((Decimal("2.5"),), Fraction(5, 2), DP, None),
    LFCase((Decimal("0.1"),), Fraction(1, 10), DP, None),
    # (Decimal, int)
    LFCase(((Decimal("0.5"), 3),), Fraction(1, 8), DP, None),
    LFCase(((Decimal("0.5"), -3),), Fraction(8), DP, None),
    # (Fraction, int)
    LFCase(((Fraction(2, 3), 2),), Fraction(4, 9), DP, None),
    LFCase(((Fraction(2, 3), -2),), Fraction(9, 4), DP, None),
    # (float, int)
    LFCase(((2.5, 2),), "6.25", DP, None),
    LFCase(((2.0, -2),), "0.25", DP, None),
    LFCase(((0.1, 2),), "0.01000000000000000111022302462", DP, None),  # inexact
    # (int, int)
    LFCase(((2, 3),), Fraction(8), DP, None),
    LFCase(((2, -3),), Fraction(1, 8), DP, None),
    LFCase(((2, 0),), Fraction(1), DP, None),
    LFCase(((2, 1),), Fraction(2), DP, None),
    # (Decimal, Fraction)
    LFCase(
        ((Decimal("16"), Fraction(1, 4)),),
        "2.000000000000000000000000000",
        DP,
        None,
    ),
    LFCase(
        ((Decimal("2"), Fraction(1, 2)),),
        "1.414213562373095048801688724",
        DP,
        None,
    ),
    # (Fraction, Fraction)
    LFCase(
        ((Fraction(1, 4), Fraction(1, 2)),),
        "0.5000000000000000000000000000",
        DP,
        None,
    ),
    LFCase(
        ((Fraction(1, 2), Fraction(1, 2)),),
        "0.7071067811865475244008443621",
        DP,
        None,
    ),
    # (float, Fraction)
    LFCase(((4.0, Fraction(1, 2)),), "2.000000000000000000000000000", DP, None),
    LFCase(
        ((0.1, Fraction(1, 2)),), "0.3162277660168379419769730258", DP, None
    ),
    # (int, Fraction)
    LFCase(((4, Fraction(1, 2)),), "2.000000000000000000000000000", DP, None),
    LFCase(((2, Fraction(1, 2)),), "1.414213562373095048801688724", DP, None),
    LFCase(((27, Fraction(2, 3)),), "9.000000000000000000000000001", DP, None),
    LFCase(((4, Fraction(-1, 2)),), "0.5000000000000000000000000000", DP, None),
    LFCase(((4, Fraction(0, 1)),), Fraction(1), DP, None),
    # 0
    LFCase(((0, 2),), Fraction(0), DP, None),
    LFCase(((0, 0),), Fraction(1), DP, None),
    LFCase(((0, Fraction(1, 2)),), Fraction(0), DP, None),
    LFCase(((0, Fraction(0, 1)),), Fraction(1), DP, None),
    LFCase(((0, -2),), None, DP, ZeroDivisionError),
    LFCase(((0, Fraction(-1, 2)),), None, DP, ZeroDivisionError),
    # 1
    LFCase(((1, 10),), Fraction(1), DP, None),
    LFCase(((1, Fraction(2, 3)),), Fraction(1), DP, None),
    # -ve
    LFCase(((-2, 2),), Fraction(4), DP, None),
    LFCase(((-2.0, 2),), "4", DP, None),
    LFCase(((-4, Fraction(1, 2)),), None, DP, InvalidOperation),
    LFCase(((-8, Fraction(1, 3)),), None, DP, InvalidOperation),
    LFCase(((Decimal("-27"), Fraction(1, 3)),), None, DP, InvalidOperation),
    # mixed
    LFCase((2, Fraction(1, 3), Decimal("3")), Fraction(2), DP, None),
    LFCase(((2.0, 2), (0.5, 1)), "2.0", DP, None),
    LFCase(
        (Fraction(1, 3), (9, Fraction(1, 2))),
        "0.9999999999999999999999999999",
        DP,
        None,
    ),
    LFCase(
        (Fraction(1, 2), (Decimal("4"), Fraction(1, 2)), (Decimal("0.5"), 1)),
        "0.5000000000000000000000000000",
        DP,
        None,
    ),
    # precision
    LFCase(((2, Fraction(1, 2)),), "1.4142136", LP, None),
    LFCase(((2, Fraction(1, 2)),), "1.414213562373095048801688724", DP, None),
    LFCase(
        ((2, Fraction(1, 2)),),
        "1.4142135623730950488016887242096980785696718753769",
        HP,
        None,
    ),
]


@pytest.mark.parametrize(
    "tc",
    [pytest.param(tc, id=f"case_{tc.products_input}") for tc in LF_CASES],
)
def test_lazy_product(
    tc: LFCase,
) -> None:
    lpwe = LazyProduct(tc.products_input)

    ctx_old = getcontext()
    with localcontext() as ctx:
        if tc.precision is not None:
            ctx.prec = tc.precision
        if tc.expected_exception:
            with pytest.raises(tc.expected_exception):
                lpwe.to_exact(ctx=ctx_old)
        else:
            result_exact = lpwe.to_exact(ctx=ctx)
            result_approx = lpwe.to_approx()
            if isinstance(tc.expected_value, Fraction):
                assert isinstance(result_exact, Fraction)
                assert result_exact == tc.expected_value
                assert result_approx == pytest.approx(float(tc.expected_value))
            elif isinstance(tc.expected_value, str):
                assert isinstance(result_exact, Decimal)
                assert str(result_exact) == tc.expected_value
                assert result_approx == pytest.approx(float(tc.expected_value))
            else:
                raise ValueError(f"unexpected {type(tc.expected_value)=}")
